from cobra.io import load_model
from csmemo.reachability import ReachabilityTester



def test_tiny_model():
    model = load_model("textbook")

    biomass = [r for r in model.reactions if r.objective_coefficient==1][0]

    target_products = [m.id for m in biomass.reactants]
    print(target_products)
    tester = ReachabilityTester(model, target_products)
    result = tester.single_gene_reachability()
    result.to_csv('reachability_frame.tsv', sep='\t')



test_tiny_model()


from os import path
from cobra.io import read_sbml_model
from cobra.manipulation import remove_genes
 
 
def rename_rxns_ids(model, rxn_rename_dict):
    for r_id_old, r_id_new in rxn_rename_dict.items():
        if r_id_old not in model.reactions:
            continue
        rxn = model.reactions.get_by_id(r_id_old)
        rxn.id = r_id_new
    model.repair()
    return model




data_dir = path.join(path.abspath(path.curdir), 'data')
sbml_fname = path.join(data_dir, 'Recon3DModel_301_no_isoforms.xml.gz')
model = read_sbml_model(sbml_fname, f_replace={})


gene_id = 'G_FOLR2'

# remove_genes(model, [gene_id])
# print("worked without renaming")



# model = read_sbml_model("Recon3DModel_301_no_isoforms.xml.gz")
rxn_rename_dict = {'FOLR2': 'FIXED_FOLR2'}
# model = rename_rxns_ids(model, rxn_rename_dict)
remove_genes(model, [gene_id])
print("worked with renaming")


import sys
import json
import os
import pandas as pd
from cobra.io import read_sbml_model
from cobra.io import load_model
from cobra.manipulation import remove_genes
from csmemo import csmbuilder



def test_tiny_model():
    model = load_model("textbook")

    gene_expression = {g.id:2 for g in  model.genes}
    gene_thresholds = {g.id:1 for g in  model.genes}

    # SUCOAS	[b0728 and b0729]
    gene_expression["b0729"] = 0

    # THD2	    [b1602 and b1603]
    gene_expression["b1603"] = 0

    # TALA	    [b2464 or b0008]
    gene_expression["b0008"] = 0
    
    # TKT1	    [b2935 or b2465]
    gene_expression["b2465"] = 0

    
    builder = csmbuilder.CSMBuilder(model, gene_expression, gene_thresholds, csm_id='toymodel', 
                                    global_lb_q_threshold=0.0, global_ub_q_threshold=1.0,  
                                    check_consistency=True, copy_model=True, debug_mode=True)

    builder.run_me_solver()
    cs_model = builder.build_csm()   

    cs_model.optimize()
    print(cs_model.summary())



def test_recon3d():
    
    data_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
    sbml_fname = os.path.join(data_dir, 'Recon3DModel_301_no_isoforms.xml.gz')
    model = read_sbml_model(sbml_fname, f_replace={})

    json_fname = os.path.join(data_dir,'MIAPACA2_PANCREAS_expression.json')
    with open(json_fname) as fh:
        gene_expression = json.load(fh)
        gene_expression = {f"G_{k}":v for k,v in gene_expression.items()}

    json_fname = os.path.join(data_dir,'expression_thresholds.json')
    with open(json_fname) as fh:
        gene_thresholds = json.load(fh)
        gene_thresholds = {f"G_{k}":v for k,v in gene_thresholds.items()}

    builder = csmbuilder.CSMBuilder(model, gene_expression, gene_thresholds, csm_id='miapaca', 
                                    global_ub_q_threshold=0.8, debug_mode=True, copy_model=False, 
                                    check_consistency=False)

    builder.run_me_solver()
    builder.build_csm()




def test_remove_gpr():
    model = read_sbml_model("data/textbook.xml.gz")
    print("=======================")
    print("Testing GPRs after removing genes")
    rxn_orig = model.reactions.ALCD2x
    print("Reaction:", rxn_orig.id)
    print(f"Genes associated to {rxn_orig.id}:", [g.id for g in rxn_orig.genes])
    print("Original reaction GPR:", rxn_orig.gene_reaction_rule)
    print("=======================")
    print("Coping model")
    model_copy = model.copy()
    print("Removing gene b1478 from model copy")
    remove_genes(model_copy, ["b1478"])
    rxn_copy = model_copy.reactions.ALCD2x
    print("=======================")
    print("Original reaction GPR (after removing copy):", rxn_orig.gene_reaction_rule)
    print("Genes associated to original reaction:", [g.id for g in rxn_orig.genes])
    print("=======================")
    print("Copied reaction GPR:", rxn_copy.gene_reaction_rule)
    print("Genes associated to original reaction:", [g.id for g in rxn_copy.genes])
    

def delete_gene_test():

    def _rename_rxns_ids(model, rxn_rename_dict):
        for r_id_old, r_id_new in rxn_rename_dict.items():
            if r_id_old not in model.reactions:
                continue
            rxn = model.reactions.get_by_id(r_id_old)
            rxn.id = r_id_new
        model.repair()
        return model

    gene_id = 'FOLR2'

    model = read_sbml_model("test/data/Recon3DModel_301_no_isoforms.xml.gz")
    try:
        remove_genes(model, [gene_id])
        print("worked (without renaming)")
    except:
        print("FAIL 1")

    model = read_sbml_model("test/data/Recon3DModel_301_no_isoforms.xml.gz")
    try:
        rxn_rename_dict = {'FOLR2': 'FIXED_FOLR2'}
        model = _rename_rxns_ids(model, rxn_rename_dict)
        remove_genes(model, [gene_id])
        print("worked (with renaming)")
    except:
        print("FAIL 2")

def main():
    # test_tiny_model()
    test_recon3d()
    # test_remove_gpr()
    # delete_gene_test()



main()


import re
import warnings
import pandas as pd
import logging
from cobra.core import Reaction
from cobra.core import DictList
from pyfastcore import Fastcore


class InfeasibleProblemError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class MetabolicTaskException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class AlreadyRegisteredError(MetabolicTaskException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class TaskRegisterError(MetabolicTaskException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def find_gene_knockout_reactions(model):
    gene_ko_reactions = {}
    for g in model.genes:
        gene_ko_reactions[g.id] = []
    
    for r in model.reactions:
        if r.gpr is None or len(r.gene_reaction_rule) == 0:
            continue
        for g in r.genes:
            if r.gpr.eval([g]):
                gene_ko_reactions[g.id].append(r.id)

    return gene_ko_reactions


class ReachabilityTester:

    def __init__(self, cobra_model, target_products,
                 epsilon=1e-4, tolerance=1e-7, 
                 flux_bound=999999., level=logging.INFO):

        self._model = cobra_model.copy()
        self._var_mapping = {}
        self._epsilon = epsilon
        self._tolerance = tolerance
        self._demands_dict = {}
        self._product_to_demands_dict = {}
        self._original_bound = {}
        self._not_reachable_products = []

        self._gene_knockout_reactions = find_gene_knockout_reactions(self._model)

        assert len(target_products) > 0

        logging.basicConfig(level=level)
        logging.info("- Creating CSMBuilder")

        
        logging.info("===========================================================")
        logging.info("Initializing Reachability Tester using")
        logging.info("Model: %s" % cobra_model.id)

        logging.info("- Nº of reactions: %i" % len(self._model.reactions))
        logging.info("- Nº of metabolites: %i" % len(self._model.metabolites))
        logging.info("- Nº of target products: %i" % len(target_products))

        if hasattr(target_products[0], 'id'):
            target_products = [m.id for m in target_products]

        for r in self._model.reactions:
            self._original_bound[r.id] = r.bounds
            if r.upper_bound > 0:
                r.upper_bound = flux_bound
            if r.lower_bound < 0:
                r.lower_bound = -flux_bound

        for m in target_products:
            m = self._model.metabolites.get_by_id(m)
            try:
                demand = self._model.add_boundary(m, type='demand')
            except (KeyError, ValueError) as e:
                demand_id = re.sub(r"^[^']*'([^']*)'.*", r"\1", e.args[0])
                demand = self._model.reactions.get_by_id(demand_id)

            self._demands_dict[demand.id] = m
            self._product_to_demands_dict[m.id] = demand.id

        self.__init_lp_problems()


        not_reachable_targets = self.test_target_products()
        if len(not_reachable_targets) > 0:
            logging.warning(" - The following target products cannot be reached")
            print(" ".join(not_reachable_targets))
            logging.warning(" - Removing compounds form the list of target products")
            for i in not_reachable_targets:
                demand = self._product_to_demands_dict[i]
                del self._demands_dict[demand]

            logging.info("Updating LP models")
            self.__init_lp_problems()

    @property
    def model(self):
        return self._model

    @property
    def epsilon(self):
        return self._epsilon

    @property
    def tolerance(self):
        return self._tolerance

    def __init_lp_problems(self):

        self._var_mapping = {}
        self._lp7 = None
        self._fva = None

        demands = list(self._demands_dict.keys())
        self._lp7 = Fastcore.create_optlang_lp7(self._model, demands, epsilon=self.epsilon)
        self._fva = Fastcore.create_optlang_fba(self._model)

        for rxn, met in self._demands_dict.items():
            z_var_id = Fastcore.AUXILIARY_VAR_PREFIX + rxn
            z_var = self._lp7.variables.get(z_var_id)
            self._var_mapping[z_var] = met

    def __run_lp7_reachability(self, reactions_to_block):
        not_produced = []
        variables_to_block = {}

        if len(reactions_to_block) > 0:
            if hasattr(reactions_to_block[0], 'id'):
                reactions_to_block = [r.id for r in reactions_to_block]

            for r in reactions_to_block:
                var = self._lp7.variables.get(r)
                variables_to_block[var] = (var.lb, var.ub)
                var.lb = 0
                var.ub = 0

        self._lp7.optimize()
        eps = 0.99 * self.epsilon
        for z_var, met in self._var_mapping.items():
            if z_var.primal - eps > self.tolerance:
                continue
            not_produced.append(met.id)

        for r in reactions_to_block:
            var = self._lp7.variables.get(r)
            (var.lb, var.ub) = variables_to_block[var]

        return not_produced

    def __run_fva_reachability(self, reactions_to_block):
        lp_model = self._fva
        not_produced = []
        if len(reactions_to_block) > 0:
            if hasattr(reactions_to_block[0], 'id'):
                reactions_to_block = [r.id for r in reactions_to_block]

        bounds_dict = {}
        for r in reactions_to_block:
            var = lp_model.variables.get(r)
            bounds_dict[var] = var.lb, var.ub
            var.lb = 0
            var.ub = 0

        objective_coefficients = {v: 0 for v in lp_model.variables}
        for r, m in self._demands_dict.items():
            var = lp_model.variables.get(r)
            objective_coefficients[var] = 1
            lp_model.objective.set_linear_coefficients(objective_coefficients)
            lp_model.update()
            lp_model.optimize()
            objective_coefficients[var] = 0

            if lp_model.status == 'infeasible':
                warnings.warn("Infeasible solution for %s" % m)

            if var.primal < self._tolerance:
                not_produced.append(m.id)

        for r in reactions_to_block:
            var = lp_model.variables.get(r)
            var.lb = bounds_dict[var][0]
            var.ub = bounds_dict[var][1]

        return not_produced

    def run_reactions_reachability(self, reactions_to_block, method='lp7'):

        if method == 'lp7':
            return self.__run_lp7_reachability(reactions_to_block)
        elif method == 'fva':
            return self.__run_fva_reachability(reactions_to_block)
        else:
            raise NotImplementedError

    def single_gene_reachability(self, gene_list=[], method='lp7', rxn_str_sep='|'):

        if len(gene_list) == 0:
            gene_list = [g.id for g in self._model.genes]

        if hasattr(gene_list[0], 'id'):
            gene_list = [g.id for g in gene_list]

    
        logging.info(f"Running reachability for {len(gene_list)} genes")
        data = []
        for g in gene_list:

            reactions_to_block = self._gene_knockout_reactions[g]
            if len(reactions_to_block) == 0:
                continue

            not_produced = self.run_reactions_reachability(reactions_to_block, method=method)
            if len(not_produced) == 0:
                continue

            reactions_strn = rxn_str_sep.join([r for r in reactions_to_block])
            for m in not_produced:
                m = self._model.metabolites.get_by_id(m)
                data.append([g, m.id, m.name, reactions_strn])

        columns = ['gene_id', 'compound_id', 'compound_name', 'reactions']
        df_reachability = pd.DataFrame(columns=columns, data=data)
        df_reachability.index.name = 'idx'

        return df_reachability

    def single_reaction_reachability(self, reaction_list=[], method='lp7', find_ums=False):

        if len(reaction_list) == 0:
            reaction_list = [r.id for r in self._model.reactions]

        if hasattr(reaction_list[0], 'id'):
            reaction_list = [r.id for r in reaction_list]

        data = []
        for r in reaction_list:

            not_produced = self.run_reactions_reachability([r], method=method)
            if len(not_produced) == 0:
                continue

            r = self._model.reactions.get_by_id(r)
            for m in not_produced:
                m = self._model.metabolites.get_by_id(m)
                data.append([r.id, m.id, m.name, r.gene_reaction_rule])

        columns = ['reaction_id', 'compound_id', 'compound_name', 'gpr']
        df_reachability = pd.DataFrame(columns=columns, data=data)
        df_reachability.index.name = 'idx'

        return df_reachability

    def test_target_products(self):
        self._not_reachable_products = self.__run_lp7_reachability([])
        return self._not_reachable_products

    @staticmethod
    def to_square_form(df, gene_col='gene_id', task_col='compound_id', value_col=None):
        if value_col is None:
            value_col = "values"
            df[value_col] = 1
            
        df_square = df.pivot_table(index=task_col, columns=gene_col, values=value_col, fill_value=0.0)

        return df_square


class TaskTester(object):

    AUXILIARY_VAR_PREFIX = "Y_var_task"
    AUXILIARY_CONS_PREFIX = "Y_const_task"

    def __init__(self, cobra_model, task_list, zero_tolerance=1e-9, info_level=logging.INFO, debug=False):

        assert len(task_list) > 0
        logging.basicConfig(level=info_level)

        self._model = cobra_model.copy()
        self._debug = debug

        logging.info("- Model: %s" % cobra_model.id)
        logging.info(f"\t|Reactions|={len(self._model.reactions)}")
        logging.info(f"\t|Metabolites|={len(self._model.metabolites)}")
        logging.info(f"\t|Genes|={len(self._model.genes)}")

        self._original_bounds = {r.id: r.bounds for r in self._model.reactions}

        for r in self._model.exchanges:
            r.bounds = (0, 0)
        
        self._tasks = DictList(task_list)
        self._solver = None
        self._sources = {}
        self._demands = {}
        self._zero_tolerance = zero_tolerance
        self._infeasible_tasks = []

        self._gene_knockout_reactions = find_gene_knockout_reactions(self._model)

        logging.info("- Creating Metabolic Task Tester")

        self.__initialize_tester()
        self.__init_lp_solver()

        logging.info(f"- Testing infeasible metabolic tasks")
        tested_taks = self.test_all_tasks()
        
        for task_id, success in tested_taks.items():
            if success:
                continue
            else:
                task = self.tasks.get_by_id(task_id)
                self._infeasible_tasks.append(task)
                self._tasks.remove(task)
        
        if len(self._infeasible_tasks) > 0:
            logging.warning(f"- {len(self._infeasible_tasks)} tasks failed test, skiping them.")

    @property
    def tasks(self):
        return self._tasks

    @property
    def infeasible_tasks(self):
        return self._infeasible_tasks

    @property
    def solver(self):
        return self._solver

    def __initialize_tester(self):
        logging.info("- Initializing steps")
        logging.info("- Registering metabolic tasks")
        logging.info(f"\t|Metabolic tasks|={len(self.tasks)}")
        fail_to_register = []
        for task in self.tasks:
            try:
                self.__register_task(task)
            except TaskRegisterError as e:
                logging.warning(f"- Cant' register task {task.id}: " + str(e))
                fail_to_register.append(task)
        
        if len(fail_to_register) > 0:
            logging.warning(f"- Removing {len(fail_to_register)} incompatible tasks from the list")
            for task in fail_to_register:
                self._tasks.remove(task)

    def __add_sink(self, metabolite, type='input'):
            if type=='input':
                sink_id = f"TASK_INPUTt_{metabolite.id}"
                coefficient = 1
            elif type == 'output':
                sink_id = f"TASK_OUTPUT_{metabolite.id}"
                coefficient = -1
            else:
                raise TaskRegisterError(f"Unknown sink type {type}")
            if sink_id in self._model.reactions:
                return sink_id
            sink = Reaction(id=sink_id, name=sink_id, lower_bound=0, upper_bound=0)
            sink.add_metabolites({metabolite: coefficient})
            self._model.add_reaction(sink)
            return sink_id
 
    def __register_task(self, task):
        logging.info(f"\tRegistering Task {task.id}")
        for m in task.inputs:
            if m not in self._model.metabolites:
                raise TaskRegisterError(f"Metabolite {m} not found in model")
            if m in self._sources:
                continue
            metabolite = self._model.metabolites.get_by_id(m)
            source_id = self.__add_sink(metabolite, type='input')
            self._sources[m] = source_id

        for m in task.outputs:
            if m not in self._model.metabolites:
                raise TaskRegisterError(f"Metabolite {m} not found in model")
            if m in self._demands:
                continue
            metabolite = self._model.metabolites.get_by_id(m)
            demand_id = self.__add_sink(metabolite, type='output')
            self._demands[m] = demand_id

    def __init_lp_solver(self):
        logging.info("- Initializing LP solver")
        self._solver = self._model.solver
        self._solver.name = 'Task Testing LP 1'
        self._objective_coefficients = {v: 0.0 for v in self._solver.variables}
        self._solver.objective.set_linear_coefficients(self._objective_coefficients)
        self._solver.objective.direction = 'max'
        self._solver.update()

    def __activate_task(self, task, activate=True):
        for m in task.inputs:
            objective_coefficients = {v: 0.0 for v in self._solver.variables}
            source_id = self._sources[m]
            var = self._solver.variables.get(source_id)
            if activate:
                var.ub = task.inputs[m][1]
                var.lb = task.inputs[m][0]
                objective_coefficients[var] = 1
            else:
                var.lb = 0.0
                var.ub = 0.0
                objective_coefficients[var] = 0.0

        for m in task.outputs:
            demand_id = self._demands[m]
            var = self._solver.variables.get(demand_id)
            if activate:
                var.ub = task.outputs[m][1]
                var.lb = task.outputs[m][0]
                objective_coefficients[var] = 1
            else:
                var.lb = 0.0
                var.ub = 0.0
                objective_coefficients[var] = 0.0
        
        self._solver.objective.set_linear_coefficients(self._objective_coefficients)
        self._solver.update()

    def test_single_task(self, task):
        if hasattr(task, 'id'):
            task = task.id
        task = self.tasks.get_by_id(task)
        self.__activate_task(task)
        self._solver.optimize()
        self.__activate_task(task, activate=False)
        return self._solver.status


    def test_all_tasks(self):
        result = {}
        for task in self.tasks:
            status = self.test_single_task(task)
            if status == 'optimal':
                result[task.id] = True
            else:
                 result[task.id] = False
        return result

    def single_gene_deletion(self, gene_list=[], rxn_str_sep='|'):

        if len(gene_list) == 0:
            gene_list = [g.id for g in self._model.genes]
        if hasattr(gene_list[0], 'id'):
            gene_list = [g.id for g in gene_list]
    
        logging.info(f"Running single gene deletion for {len(gene_list)} genes on {len(self.tasks)} tasks")
        data = []
        for g in gene_list:
            
            if self._debug:
                logging.info(f"Testing gene {g}")

            reactions_to_block = self._gene_knockout_reactions[g]
            if len(reactions_to_block) == 0:
                continue

            if self._debug:
                logging.info(f"Disbling reactons: {reactions_to_block}")

            bounds = {}
            for r in reactions_to_block:
                var = self._solver.variables.get(r)
                bounds[r] = (var.lb, var.ub)
                var.lb = 0.0
                var.ub = 0.0          
            self._solver.update()

            result = self.test_all_tasks()

            for r in reactions_to_block:
                var = self._solver.variables.get(r)
                var.ub = bounds[r][1] 
                var.lb = bounds[r][0]
            self._solver.update()

            failed_task = [k for k,v in result.items() if not v]

            if len(failed_task) == 0:
                continue

            reactions_strn = rxn_str_sep.join([r for r in reactions_to_block])
            for t_id in failed_task:
                task = self.tasks.get_by_id(t_id)
                data.append([g, task.id, task.system, task.subsystem, reactions_strn])

        columns = ['gene_id', 'task_id', 'system', 'subsystem', 'reactions']
        df_reachability = pd.DataFrame(columns=columns, data=data)
        df_reachability.index.name = 'idx'

        return df_reachability

    def __init_milp(self):
        logging.info("- Initializing MILP model")
        optlang_interface = self._model.solver.interface
        self._milp_model = Fastcore.create_optlang_fba(self._model)
        self._milp_model.name = 'Task Testing MILP 1'

        self._objective_coefficients = {v: 0.0 for v in self._milp_model.variables}
        const_to_add = []
        for task in self.task_list:
            y_var_id = f"{TaskTester.AUXILIARY_VAR_PREFIX}_{task.id}"
            y_var = optlang_interface.Variable(y_var_id, lb=0, ub=1, type='binary')
            self._milp_model.add(y_var)
            self._objective_coefficients[y_var] = 1.0
            self._task2vars[task.id] = y_var

            for m, source in task.sources.items():
                lb, ub = task.get_bounds(m)
                v_var = self._milp_model.variables.get(source.id)
                v_var.lb = 0.0
                v_var.ub = 1000.0

                y_constraint_id = f"{TaskTester.AUXILIARY_CONS_PREFIX}_{task.id}_{source.id}_LHS"
                constraint = optlang_interface.Constraint(v_var - y_var * lb, lb=0, name=y_constraint_id)
                const_to_add.append(constraint)

                y_constraint_id = f"{TaskTester.AUXILIARY_CONS_PREFIX}_{task.id}_{source.id}_RHS"
                constraint = optlang_interface.Constraint(y_var * ub - v_var, lb=0, name=y_constraint_id)
                const_to_add.append(constraint)
            
            for m, demand in task.demands.items():
                lb, ub = task.get_bounds(m)
                v_var = self._milp_model.variables.get(demand.id)
                v_var.lb = 0.0
                v_var.ub = 1000.0

                y_constraint_id = f"{TaskTester.AUXILIARY_CONS_PREFIX}_{task.id}_{demand.id}_LHS"
                constraint = optlang_interface.Constraint(v_var - y_var * lb, lb=0, name=y_constraint_id)
                const_to_add.append(constraint)

                y_constraint_id = f"{TaskTester.AUXILIARY_CONS_PREFIX}_{task.id}_{demand.id}_RHS"
                constraint = optlang_interface.Constraint(y_var * ub - v_var, lb=0, name=y_constraint_id)
                const_to_add.append(constraint)
        
        self._milp_model.add(const_to_add)
        self._milp_model.objective.set_linear_coefficients(self._objective_coefficients)
        self._milp_model.objective.direction = 'max'
        self._milp_model.update()
    def __evaluate_tasks_milp(self, task_list):
        milp_model = self._milp_model
        if len(task_list) == 0:
            return
        # Setting objective coefficients to zero
        self._objective_coefficients = {x: 0.0 for x in milp_model.variables}
        
        # Setting coefficients of binary var correspondint to the tested tasks to one
        for task in task_list:
            y_var = self._task2vars[task.id]
            self._objective_coefficients[y_var] = 1.0
        
        # Updating milp model
        milp_model.objective.set_linear_coefficients(self._objective_coefficients)
        milp_model.update()

        task_dict = {task.id: task for task in task_list}
        succeed_tasks = set()
        failed_tasks = set()
        untested_taks = set([task.id for task in task_list])
        while len(untested_taks) > 0:
            milp_model.optimize()
            if milp_model.status == 'infeasible':
                raise InfeasibleProblemError()

            for task_id in untested_taks:
                y_var = self._task2vars[task_id]
                if  abs(1 - y_var.primal) < self.zero_tolerance:
                    succeed_tasks |= set( [task_id] )
                    
            if len(untested_taks & succeed_tasks) > 0:
                untested_taks -= succeed_tasks
            else:
                failed_tasks = untested_taks
                untested_taks = set()
        
        return [task_dict[i] for i in failed_tasks]
    


class MetabolicTask(object):

    TASK_PREIX = "TASK"
    
    def __init__(self, id, inputs, outputs, system="", subsystem="", 
                 description="", annotations={}, should_fail=False):
        
        self._id = str(id)
        self._system = system
        self._subsystem = subsystem
        self._description = description
        self._annotations = annotations
        self._inputs = inputs 
        self._outputs = outputs

        self._should_fail = should_fail
        self._model = None
        self._sources = None
        self._demands = None

        self._original_objective = {}
        
    @property
    def id(self):
        return self._id
        
    @property
    def system(self):
        return self._system
    
    @property
    def subsystem(self):
        return self._subsystem

    @property
    def inputs(self):
        return self._inputs
    
    @property
    def outputs(self):
        return self._outputs

    @property
    def sources(self):
        return self._sources
    
    @property
    def demands(self):
        return self._demands

    @property
    def is_registered(self):
        return self._model is not None

    def get_bounds(self, m_id):
        if m_id in self._inputs:
            return self._inputs[m_id]
        if m_id in self._outputs:
            return self._outputs[m_id]
        else:
            return None
           
    def to_dict(self):
        return {
                "id": self._id,
                "inputs": self._inputs,
                "outputs": self._outputs,
                "should_fail": self._should_fail,
                "system": self._system,
                "subsystem": self._subsystem,
                "description": self._description,
                "annotations": self._annotations
            }
    
    def to_frame(self):
        
        columns = ["id", "metabolite",  "lower_bound", "upper_bound", "is_output",
                   "should_fail", "system", "subsystem", "description", "annotations"]    
        
        data = []
        for m_id in self.inputs:
            row = {}
            row["id"] = self._id
            row["metabolite"] = m_id
            row["lower_bound"] = self.inputs[m_id][0]
            row["upper_bound"] = self.inputs[m_id][1]
            row["is_output"] = False
            row["should_fail"] = self._should_fail
            row["system"] = self._system
            row["subsystem"] = self._subsystem
            row["description"] = self._description
            for k,v in self._annotations.items():
                row[k.lower()] = v
            data.append(row)
        for m_id in self.outputs:
            row = {}
            row["id"] = self._id
            row["metabolite"] = m_id
            row["lower_bound"] = self.outputs[m_id][0]
            row["upper_bound"] = self.outputs[m_id][1]
            row["is_output"] = True
            row["should_fail"] = self._should_fail
            row["system"] = self._system
            row["subsystem"] = self._subsystem
            row["description"] = self._description
            for k,v in self._annotations.items():
                row[k.lower()] = v
            data.append(row)
        
        columns = list(row.keys())
        
        df = pd.DataFrame(data, columns=columns, )
        return df.set_index(['id', 'metabolite'])



    def remove(self):
        if self._model is None:
            return
        self._model.remove_reactions([r for r in self._sources])
        self._sources = None
        self._model.remove_reactions([r for r in self._demands])
        self._demands = None
        self._model = None

    def activate(self):
        if not self.is_registered:
            return

        self._original_objective = {}
        for r in self._model.reactions:
            self._original_objective[r.id] = r.objective_coefficient
            if r.objective_coefficient == 0:
                continue
            r.objective_coefficient = 0
        
        for m in self.inputs:
            self.sources[m].bounds = self.inputs[m]
            self.sources[m].objective_coefficient = 1
        
        for m in self.outputs:
            self.demands[m].bounds = self.outputs[m]
            self.demands[m].objective_coefficient = 1

    def deactivate(self):
        if not self.is_registered:
            return
        for m in self.inputs:
            self.sources[m].bounds = (0.0, 0.0)
        for m in self.outputs:
            self.demands[m].bounds = (0.0, 0.0)
        
        for r in self._original_objective:
            rxn = self._model.reactions.get_by_id(r)
            rxn.objective_coefficient = self._original_objective[r]


import json
import pandas as pd


def read_json(json_fname):
    with open(json_fname) as fh:
        result = json.load(fh)
    return result


def read_genes_table(fname, index='hgnc_id'):
    df = pd.read_csv(fname, delimiter='\t')
    df.set_index(index, inplace=True)
    return df



import enum
import re
import logging
import boolean
import numpy as np
import pandas as pd
from cobra.core.gene import GPR
from cobra.manipulation import remove_genes

from pyfastcore import Fastcore

from ast import (
    And,
    BoolOp,
    Expression,
    Name,
    Or,
)



GPR_LOGIC_OR_MAP = ('max', 'sum')
ME_METHODS = ('fastcore', )


def _get_enzymes_recursive(expression, or_operator="|", and_operator="&"):
    list_of_enzymes = []
    if expression.isliteral:
        list_of_enzymes.append(str(expression))
    else:
        expression = expression.simplify()
        if expression.operator == or_operator:
            for arg in expression.args:
                if arg.isliteral:
                    list_of_enzymes.append([str(arg)])
                else:
                    enzymes = _get_enzymes_recursive(arg)
                    list_of_enzymes.extend(enzymes)
        elif expression.operator == and_operator:
            enzymes = []
            simple = True
            for arg in expression.args:
                subunits = _get_enzymes_recursive(arg)
                if len(subunits) > 1:
                    simple = False
                    enzymes = [enzymes + s for s in subunits]
                else:
                    enzymes.append(subunits[0])
            if simple:
                list_of_enzymes.append(enzymes)
            else:
                list_of_enzymes.extend(enzymes)
        else:
            raise Exception("ERROR IN EXPRESSION: Unknown operator/literar:", str(expression))

    return list_of_enzymes


def get_enzymes(gene_reaction_rule):
    dash_found = False
    if re.search("-", gene_reaction_rule):
        gene_reaction_rule = re.sub("-", "_", gene_reaction_rule)
        dash_found = True

    algebra = boolean.BooleanAlgebra()
    expression = algebra.parse(gene_reaction_rule, simplify=True)
    if expression.isliteral:
        list_of_enzymes = [[str(expression)]]
    else:
        expression = expression.simplify()
        list_of_enzymes = _get_enzymes_recursive(expression)
        for enzyme in list_of_enzymes:
            wrong_parse = False
            for subunit in enzyme:
                if isinstance(subunit, str):
                    continue
                wrong_parse = True
                break
        if wrong_parse:
            expression = expression.distributive()
            list_of_enzymes = _get_enzymes_recursive(expression)
    if dash_found:
        for enzyme in list_of_enzymes:
            for i,gene in enumerate(enzyme):
                enzyme[i] = re.sub("_", "-", gene)           
    list_of_enzymes = tuple([tuple(sorted(i)) for i in list_of_enzymes])
    return list_of_enzymes


def rename_element_ids(model, rename_dict, type='reaction'):
    if type == 'reaction':
        element_list = model.reactions
    for id_old, id_new in rename_dict.items():
        if id_old not in element_list:
            continue
        el = element_list.get_by_id(id_old)
        el.id = id_new
    model.repair()
    return model


class UnsupportedGPROperator(Exception):
    def __init__(self, message="Supported operators for GPR logic \"or\" are \"max\" or \"sum\""):
        self.message = message
        super().__init__(self.message)


class UnknownMEMethod(Exception):
    message = "Supported ME methods include:", " ".join(ME_METHODS)
    def __init__(self, message=message):
        self.message = message
        super().__init__(self.message)


class CSMBuilder:
    
    def __init__(self, cobra_model, gene_expression, gene_thredholds, csm_id=None,
                 me_method='fastcore', gpr_or='max', fill_missing_gene=0.0, copy_model=True,
                 global_lb_q_threshold=0.25, global_ub_q_threshold=0.7, base_core=[],
                 penalty_weights={0.75:0, 0.5:-1, 0.25:-10, 0.1:-100}, epsilon=1e-7,
                 level=logging.INFO, init_me_solver=True, **kwargs):
        """

        Parameters
        ----------
        cobra_model : cobra Model 
            A model to be use as universal or reference
        gene_expression: dict 
            A dictionarywith gene IDs (present in the model) and expression values
        gene_thredholds: dict
            A dictionary with gene IDs (present in the model) and thresholds to binarize the expression
        fill_missing_gene: float
            Default expression to assign to genes with unknown values
        me_method: String
            The model extraction method used to create a context-specific model
        """

        logging.basicConfig(level=level)
        logging.info("- Creating CSMBuilder")

        if me_method not in ME_METHODS:
            raise UnknownMEMethod("")
        self._me_method = me_method

        if gpr_or not in GPR_LOGIC_OR_MAP:
            raise UnsupportedGPROperator()
        self._gpr_or = gpr_or

        if copy_model:
            self._model = cobra_model.copy()
        else:
            self._model = cobra_model

        self._original_bounds = {}
        for r in self._model.reactions:
            self._original_bounds[r.id] = r.bounds

        self._csm_id = csm_id
        if self._csm_id is None:
            self._csm_id = self._model.id

        assert fill_missing_gene >= 0
        self._fill_missing_gene = fill_missing_gene

        self._gene_expression = dict()
        self.__consolidate_genes(gene_expression)

        assert global_lb_q_threshold >= 0.0
        assert global_lb_q_threshold <= 1.0 
        assert global_ub_q_threshold >= 0.0
        assert global_ub_q_threshold <= 1.0
        assert global_lb_q_threshold < global_ub_q_threshold
        
        self._global_lb_threshold = 0.0
        self._global_lb_threshold = max(gene_expression.values())
        
        self.__calculate_global_thresholds(global_lb_q_threshold, global_ub_q_threshold)

        self._gene_thresholds = dict()
        self.__consolidate_thresholds(gene_thredholds)

        self._gene_confidences = {}
        self.__binarize_gene_expression()

        self._reaction_expression = {}
        self.__map_gene_expression()

        self._base_core = []
        self.__consolidate_base_core(base_core)

        self._inferred_core = []
        self.__calculate_reaction_core()

        self._penalty_weights = penalty_weights
        self._penalties = {}
        self.__calculate_penalties()

        self._epsilon = epsilon

        self._kwargs = kwargs
        
        self._consistent_subnetwork = []
        self._cs_model = None

        self._me_solver = None
        if init_me_solver:
            self.__init_me_solver()

    #############
    # Properties
    #############

    @property
    def model(self):
        return self._model

    @property
    def csm_id(self):
        return self._csm_id
    
    @csm_id.setter
    def csm_id(self, csm_id):
        self._csm_id = csm_id

    @property
    def gene_expression(self):
        return self._gene_expression

    @property
    def gene_thresholds(self):
        return self._gene_thresholds

    @property
    def gene_confidences(self):
        return self._gene_confidences
    
    @property
    def reaction_expression(self):
        return self._reaction_expression

    @property
    def gpr_or(self):
        return self._gpr_or

    @property
    def base_core(self):
        return self._base_core

    @property
    def inferred_core(self):
        return self._inferred_core

    @property
    def core_reactions(self):
        core_reactions = self.base_core + self.inferred_core
        return frozenset(core_reactions)

    @property
    def penalties(self):
        return self._penalties

    @property
    def epsilon(self):
        return self._epsilon

    @property
    def cs_model(self):
        if self._cs_model is None:
            logging.warning("- CS models has not been built. First run CSMBuilder.build_csm()")
        return self._cs_model


    ##################
    # Private methods
    ##################

    def __consolidate_genes(self, gene_expression):
        """
        Find genes from the model that are  not present in the gene expression dict
        and fill the expression values of missing using the fill_missing_gene parameter
        """
        logging.info("* Consolidating genes")
        logging.info(f"\t- Total genes in model: {len(self._model.genes)}")
        missing = 0
        for g in self.model.genes:
            if g.id in gene_expression:
                self._gene_expression[g.id] = gene_expression[g.id]
            else:
                self._gene_expression[g.id] = self._fill_missing_gene
                missing += 1
        logging.info(f"\t- Total missing genes found: {missing}")
        logging.info(f"\t- Filling missing genes with: {self._fill_missing_gene}")
    
    def __calculate_global_thresholds(self, lb_quantile, ub_quantile):
        s = pd.Series(self._gene_expression)
        logging.info(f"* Calculating global lb and lb threshold using quantiles {lb_quantile} and {ub_quantile} on gene expression")
        global_lb_thr = s[s>0].quantile(lb_quantile)
        self._global_lb_threshold = global_lb_thr
        logging.info(f"\t- Global lb threshold value: {global_lb_thr}")
        global_ub_thr = s[s>0].quantile(ub_quantile)
        self._global_ub_threshold = global_ub_thr
        logging.info(f"\t- Global ub threshold value: {global_ub_thr}")
        
    def __consolidate_thresholds(self, gene_thredholds):
        """
        Find threshold from the model that are not present in the gene threshold dict
        and fill the expression values of missing using the fill_missing_gene parameter
        """
        logging.info("* Consolidating thresholds")
        logging.info(f"\t- Total genes in model: {len(self.model.genes)}")
        missing = 0
        for g in self.model.genes:
            if g.id in gene_thredholds:
                if gene_thredholds[g.id] < self._global_lb_threshold:
                    self._gene_thresholds[g.id] = self._global_lb_threshold
                elif gene_thredholds[g.id] > self._global_ub_threshold:
                    self._gene_thresholds[g.id] = self._global_ub_threshold
                else:
                    self._gene_thresholds[g.id] = gene_thredholds[g.id]
            else:
                self._gene_thresholds[g.id] = self._global_lb_threshold
                missing += 1
        logging.info(f"\t- Total missing genes found: {missing}")
        logging.info(f"\t- Filling missing genes with: {self._fill_missing_gene}")

    def __binarize_gene_expression(self):
        """
        Binarize gene expression values into {0,1} appling gene specific thredholds
        """
        logging.info("* Binarizing gene expression")
        total_active = 0
        total_inactive = 0
        for g in self._model.genes:
            confidence = self._gene_expression[g.id] >= self._gene_thresholds[g.id]
            self._gene_confidences[g.id] = confidence
            if confidence:
                total_active += 1
            else:
                total_inactive += 1

        logging.info(f"\t- Total genes {len(self._model.genes)}")
        logging.info(f"\t- Total active genes {total_active}")
        logging.info(f"\t- Total inactive genes {total_inactive}")

    def __map_gene_expression(self):
        """
        Map gene expression (continous) values thruoug GPRs into reaction expression values
        """
        aux_dict = {}
        for r in self.model.reactions:
            if r.gpr is None or len(r.gene_reaction_rule) == 0:
                self._reaction_expression[r.id] = np.nan
                continue
            if r.gene_reaction_rule not in aux_dict:
                values_dict = {g.id:self.gene_expression[g.id] for g in r.genes}
                result = self.safe_eval_gpr(r.gpr.body, values_dict, or_func=self.gpr_or)
                aux_dict[r.gene_reaction_rule] = result
            self._reaction_expression[r.id] = aux_dict[r.gene_reaction_rule]

    def __consolidate_base_core(self, base_core):
        logging.info("* Consolidating user provided core")
        if type(base_core) is str:
            base_core = [base_core]
        
        total_excluded = 0
        for r in base_core:
            if r in self._model.reactions:
                self._base_core.append(r)
            else:
                total_excluded += 1
                logging.warning(f"\t- User provided base core reaction {r} not in model.")
        if total_excluded > 0:
            logging.warning(f"\t- Excluding {len(r)} user provided base core reactions not founded in model {self._model.id}.")
        logging.info(f"\t- Base core size: {len(self.base_core)} reactions")

    def __calculate_reaction_core(self):
        logging.info("* Calculating gene inferred core reactions")
        for r in self.model.reactions:
            if r.gpr is None or len(r.gene_reaction_rule) == 0:
                continue
            inactive_genes = {g.id for g in r.genes if not self._gene_confidences[g.id]}
            if r.gpr.eval(inactive_genes):
                self._inferred_core.append(r.id)
        logging.info(f"\t- Inferred core size: {len(self.inferred_core)} reactions")
           
    def __calculate_penalties(self):
        logging.info(f"* Calculating reaction penalties using quantiles")
        reaction_expression = pd.Series(self._reaction_expression)
        non_zero_expression = reaction_expression[reaction_expression>0]
        self._penalties = pd.Series(0.0, index=reaction_expression.index)
        for q,p in self._penalty_weights.items():
            thr = non_zero_expression.quantile(q)
            mask = reaction_expression < thr
            self._penalties[mask] = p
        self.penalties[self.core_reactions] = 0.0

    def __init_me_solver(self):
        if self._me_method == 'fastcore':
            self._me_solver = Fastcore(self.model, self.core_reactions, penalties=self.penalties, copy_model=False, **self._kwargs)
        else:
            raise UnknownMEMethod()

    def __correct_thresholds(self, lp_fname=None):
        if self._cs_model is None:
            return False

        logging.info(f"* Consolidating gene activity for non-core reactions")
        consistent_subnetwork = set(self._consistent_subnetwork)
        non_core_reactions = consistent_subnetwork - self.core_reactions
        gene_list = []
        gpr_list = []
        for r in non_core_reactions:
            r = self._cs_model.reactions.get_by_id(r)
            if r.gpr is None or len(r.gene_reaction_rule) == 0:
                continue
            gpr_list.append(r.gene_reaction_rule)
            gene_list.extend([g.id for g in r.genes])
        
        logging.info(f"\t- {len(gpr_list)} non-core reaction requires an update in their gene context.")
        gpr_list = set(gpr_list)
        gene_list = set(gene_list)

        logging.info(f"\t- Processing {len(gpr_list)} unique GPR containing {len(gene_list)} genes.")

        gene_weights = {}
        for g in gene_list:
            weight = self.gene_thresholds[g] - self.gene_expression[g] 
            gene_weights[g] = weight
        
        optlang_interface = self.cs_model.solver.interface
        logging.info(f"\t- Creating MinGene MILP problem")
        logging.info(f"\t- Finding minimal gene context compatible with the non-core-reactions")
        milp_model = CSMBuilder.create_mingene_milp(optlang_interface, gpr_list, gene_weights=gene_weights)
        if lp_fname is not None:
            with open(str(lp_fname), 'w') as fh:
                fh.write(milp_model.to_lp())
        
        milp_model.optimize()
        if milp_model.status != 'optimal':
            logging.warning(f"\t- Min Gene problem could not")
            return False

        logging.info(f"\t- Optimal solution found for MinGene problem!")
        new_active_genes = [k for k,v in milp_model.primal_values.items() 
                                if v >= 0.9999 and k in gene_list]
        logging.info(f"\t- Context adjustment requires to change the status of {len(new_active_genes)} genes to active")
        logging.info(f"\t- Adjusting context be reducing gene thresholds")
        for g in new_active_genes:
            # Skiping already active genes
            if self._gene_confidences[g]:
                continue
            if self._gene_expression[g] >= self._gene_thresholds[g]:
                continue
            relaxed_threshold = self._gene_expression[g] - self.epsilon
            relaxed_threshold = max(0, relaxed_threshold)
            self._gene_thresholds[g] = relaxed_threshold
            self._gene_confidences[g] = True
        self.__binarize_gene_expression()
        
        return True

    ################
    # Pulic methods
    ################

    def run_me_solver(self):
        self._cs_model = None
        self._consistent_subnetwork = []
        if self._me_method == 'fastcore':
            logging.info(f"Running {self._me_method} model extraction method")
            self._me_solver.fast_core()
            self._consistent_subnetwork = self._me_solver.consistent_subnetwork
        else:
            raise UnknownMEMethod("")
        logging.info(f"Model extraction method finish!")
    
    def build_csm(self, consolidate_genes=True, save_mingene=None):
        
        if len(self._consistent_subnetwork) == 0:
            logging.warning("ME solver has not be run, running it now")
            self.run_me_solver()
        
        logging.info(f"Creating context-model")
        self._cs_model = self._model.copy()
        self._cs_model.id = self.csm_id
        to_remove = []
        for r in self._cs_model.reactions:
            if r.id in self._consistent_subnetwork:
                continue
            to_remove.append(r.id)
        self._cs_model.remove_reactions(to_remove)

        inactive_metabolites = [m for m in self._cs_model.metabolites if len(m.reactions) == 0]
        self._cs_model.remove_metabolites(inactive_metabolites)
        self._cs_model.repair()

        # Restoring de original bounds (Fastcore relax them)
        for r in self._cs_model.reactions:
            r.bounds = self._original_bounds[r.id]
        
        if consolidate_genes:
            self.__correct_thresholds(lp_fname=save_mingene)
            genes_to_remove = [g for g,conf in self.gene_confidences.items() 
                                        if conf == False and g in self._cs_model.genes]
            genes_to_remove = list(set(genes_to_remove))
            
            # Fix needed to avoid the error that cobra.manipulation.remove_genes raises when trying 
            # to remove a gene which has the same ID than a reactions (e.g. 'ADA', 'AGPAT1', 'RPE')
            # The fix consists in temporary renaming conflicting reactions before removing the genes
            # and then restor the original reaction's ID back.
            rename_fix_dict = {}
            rename_restore_dict = {}
            for r in self._cs_model.reactions:
                if r.id not in self._cs_model.genes:
                    continue
                fixed_r_id = "FIXED_" + r.id
                rename_fix_dict[r.id] = fixed_r_id
                rename_restore_dict[fixed_r_id] = r.id

            rename_element_ids(self._cs_model, rename_fix_dict)
            logging.info(f"\t- Pruning {len(genes_to_remove)} inactive genes from the context-specific model")
            remove_genes(self._cs_model, genes_to_remove)
            self._cs_model.repair()
            rename_element_ids(self._cs_model, rename_restore_dict)

        return self._cs_model


    @staticmethod
    def create_mingene_milp(optlang_interface, gpr_list, gene_weights={}, default_weight=1.0):
        """
        Create an optlang MILP problem for finding a subset of genes that minimize a linear 
        cost function on the genes, subject to a set of MILP constraints constructed from
        the set of boolean expression that represent Gene-Protein-Reaction rules (gpr_list)
        
        Parameters
        ----------
        optlang_interface: module
            an optlang interface used to create the Model
        gpr_list: list
            a list of strings representing gene-protein-rules as boolean
            expression such as ['(A and B) or (A and C)', 'A', 'A or C', 'E and D']
        gene_weights: dict, optional
            a dict of gene ids and the weights used to create the objective function
        default_weight: float, optional
            defulat value to use as the objective coefficinet of genes not found in gene_weights 

        """

        gene_list = []
        complex_list = []
        grp_enzymes_dict = {}
        for gpr in gpr_list:
            enzyme_list = get_enzymes(gpr)
            grp_enzymes_dict[gpr] = enzyme_list
            for enzyme in enzyme_list:
                gene_list.extend(enzyme)
                if len(enzyme) > 1:
                    complex_list.append(enzyme)

        gene_list = sorted(set(gene_list))
        complex_list = sorted(set(complex_list), key=lambda x: len(x))

        if len(gene_weights) != len(gene_list):
            for g in gene_list:
                if g in gene_weights:
                    continue
                gene_weights[g] = default_weight

        lp_model = optlang_interface.Model()
        lp_model.name = 'MinGene'
        var_dict = {}

        # Adding gene variables
        for g in gene_list:
            x_var = optlang_interface.Variable(g, lb=0, ub=1, type='continuous')
            var_dict[g] = x_var
            lp_model.add(x_var)
        lp_model.update()

        # Adding enzyme complex binary variables and constraints
        for i,enzyme in enumerate(complex_list):
            c_id = f"Complex_{i+1}"
            y_var = optlang_interface.Variable(c_id, lb=0, ub=1, type='binary')
            var_dict[enzyme] = y_var
            lp_model.add(y_var)
            expr = None
            for g in enzyme:
                x_var = var_dict[g]
                if expr is None:
                    expr = x_var
                else:
                    expr += x_var
            expr -= len(enzyme) * y_var
            constr = optlang_interface.Constraint(expr, lb=0, ub=len(enzyme))
            lp_model.add(constr)
        lp_model.update()

        # Adding GPR constraints
        for gpr, enzyme_list in grp_enzymes_dict.items():
            expr = None
            for enzyme in enzyme_list:
                if len(enzyme) == 1:
                    enzyme = enzyme[0]
                var = var_dict[enzyme]
                if expr is None:
                    expr = var
                else:
                    expr += var
            cons_name = gpr.replace(" ","")
            cons_name = cons_name.replace("and","&")
            cons_name = cons_name.replace("or","|")    
            constr = optlang_interface.Constraint(expr, lb=1)
            lp_model.add(constr)
        lp_model.update()

        objective_coefficients = {lp_model.variables.get(g): gene_weights[g]
                                  for g in gene_list}

        lp_model.objective.set_linear_coefficients(objective_coefficients)
        lp_model.objective.direction = 'min'
        lp_model.update()
        return lp_model

    @staticmethod
    def safe_eval_gpr(expr, values_dict, or_func='max'):
        if isinstance(expr, Expression):
            return CSMBuilder.safe_eval_gpr(expr.body, values_dict)
        elif isinstance(expr, Name):
            fgid = expr.id
            if fgid not in values_dict:
                return 0
            return values_dict[fgid]
        elif isinstance(expr, BoolOp):
            op = expr.op
            if isinstance(op, Or):
                if or_func == 'max':
                    return max(CSMBuilder.safe_eval_gpr(i, values_dict) for i in expr.values)
                elif or_func == 'sum':
                    return sum(CSMBuilder.safe_eval_gpr(i, values_dict) for i in expr.values)
                else:
                    raise UnsupportedGPROperator
            elif isinstance(op, And):
                return min(CSMBuilder.safe_eval_gpr(i, values_dict) for i in expr.values)
            else:
                raise TypeError("unsupported operation " + op.__class__.__name__)
        elif expr is None:
            return 0
        else:
            raise TypeError("unsupported operation  " + repr(expr))

    

    



import unittest
from ..consistency_analysis import UmFinder
from cobra.test import create_test_model


class TestUmFinder(unittest.TestCase):

    def __init__(self):
        self._model = create_test_model('iJO1366')

    def setUp(self):
        pass

    def test_model(self):
        self.assertTrue(len(self._model.reactions) == 2583)
        self.assertTrue(len(self._model.metabolites) == 1805)

    def test_umfinder(self):
        um_finder = UmFinder(self._model)
        self.assertTrue(len(um_finder.blocked_reactions) == 226)
        self.assertTrue(len(um_finder.gap_metabolites) == 203)
        self.assertTrue(len(um_finder.unconnected_modules) == 105)


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
import argparse
import os
from os import path

import pandas as pd
from cobra.io import read_sbml_model, write_sbml_model
from cobra.util import solvers
from corda import CORDA
from pyfastcore import Fastcore

from csmemo.cmds.io import read_json
from csmemo.cmds.manipulate import set_medium

MEM_METHODS = ['corda', 'fastcore']
SOLVERS = list(solvers.keys())
if 'cglpk' in SOLVERS:
    SOLVERS.remove('cglpk')
    SOLVERS.append('glpk')


def create_parser():
    parser = argparse.ArgumentParser(description='Build a context specific genome-scale metabolic model.')
    parser.add_argument('sbml_fname', action="store", help='SBML file to use a the model reference')
    parser.add_argument('json_confidences', action="store", help='CSV file storing the gene confidences')
    parser.add_argument('model_id', action="store", help='Model id for the builded CSM')

    parser.add_argument('--media', action="store", dest="json_exchanges", default=None,
                        help='JSON file storing the exchange fluxes bounds')

    parser.add_argument('--method', action='store', dest='method', choices=MEM_METHODS, default='corda',
                        help='Method used to build the CSM: corda / fastcore')

    parser.add_argument('--force-biomass', action='store_true', dest='force_biomass',
                        help='Biomass reactions are set to have the highest confidence')

    parser.add_argument('--solver', action="store", dest="solver", choices=SOLVERS,
                        default='glpk', help='LP solver to perform optimizations')

    parser.add_argument('--write-confidences', action='store_true', dest='write_confidences',
                        help='Store computed reaction confidences in CSV format')

    parser.add_argument('--default-confidence', action='store', dest='default_confidence', default=0.0, type=float,
                        help='Confidence value used by default when a model\'s gene is not found in the confidence input dict')

    parser.add_argument('--out', action="store", dest="output_folder", default=".",
                        help='Output folder to store the builded CSM')

    parser.add_argument('--use-fbc', action='store', dest='use_fbc', default=True, type=bool,
                        help='Write SBML using FBC package')                        


    curdir = path.dirname(path.realpath(__file__))
    fastcore_params_fname = path.join(curdir, "config/fastcore.json")
    parser.add_argument('--fastcore-params', action="store", dest="fastcore_params_fname",
                        default=fastcore_params_fname, help='JSON file with fastcore parameters')

    return parser

   


def generate_model_id(args, sep="_"):
    if args.force_biomass:
        model_id = sep.join([args.model_id, args.method, args.solver, "biomass"])
    else:
        model_id = sep.join([args.model_id, args.method, args.solver])

    return model_id


def generate_output_fname(args, file_type):
    fname = generate_model_id(args)
    fname = ".".join([fname, file_type])
    fname = os.path.join(args.output_folder, fname)
    return fname





def main():
    parser = create_parser()
    args = parser.parse_args()
    biomass_reaction = "biomass_reaction"

    assert os.path.isfile(args.sbml_fname)
    assert os.path.isfile(args.json_confidences)
    assert os.path.isdir(args.output_folder)

    # Reading Reference Genome-Scale Model
    print("Reading SBML Model from %s:" % args.sbml_fname, end=" ")
    model = read_sbml_model(args.sbml_fname)
    print("OK!")
    # Setting optimization solver
    print("Setting optimization solver %s:" % args.solver, end=" ")
    model.solver = args.solver
    print("OK!")
    # Reading DataFrame including gene confidence
    print("Reading gene confidence from %s:" % args.json_confidences, end=" ")
    gene_confidence_omic = read_json(args.json_confidences)
    
    gene_confidence = {}
    for g in model.genes:
        if g.id in gene_confidence_omic:
            gene_confidence[g.id] = gene_confidence_omic[g.id]
        else:
            print(f"Warning gene {g.id} not found in the confidences dict, using default confidence ({args.default_confidence})")
            gene_confidence[g.id] = args.default_confidence

    print("OK!")
    print("Computing reactions confidences:", end=" ")
    reaction_confidences = create_reactions_confidences(model, gene_confidence)
    print("OK!")

    # If True: biomass reactions are set to have the highest confidence
    if args.force_biomass:
        reaction_confidences[biomass_reaction] = 3
        print("Adding biomass reactions %s to the core set: OK!" % biomass_reaction)

    if args.json_exchanges:
        print("Reading exchange fluxes bounds: %s:" % args.json_exchanges, end=" ")
        media_dict = read_json(args.json_exchanges)
        print("OK!")
        print("Setting exchange fluxes bounds:", end=" ")
        set_medium(model, media_dict, inplace=True)
        print("OK!")

    print("Building context-specific model using \"%s\":" % args.method)
    if args.method == "corda":
        corda = CORDA(model, reaction_confidences)
        corda.build()
        cs_model = corda.cobra_model()

    elif args.method == "fastcore":

        print("Reading fastcore parameters: %s:" % args.fastcore_params_fname, end=" ")
        fastcore_params = read_json(args.fastcore_params_fname)
        print("OK!")
        fc_builder = create_fastcore(model, reaction_confidences, fastcore_params)
        fc_builder.fast_core()
        cs_model = fc_builder.build_context_specific_model()

    print("OK!")

    # Removing isolated metabolites
    isolated_metabolites = [m for m in cs_model.metabolites if len(m.reactions) == 0]
    cs_model.remove_metabolites(isolated_metabolites)

    isolated_genes = [g for g in cs_model.genes if len(g.reactions) == 0]
    for g in isolated_genes:
        cs_model.genes.remove(g)

    model_id = generate_model_id(args)
    cs_model.id = model_id
    cs_model.repair()


    sbml_output = generate_output_fname(args, 'xml')
    print("Writing context-specific model to %s:" % sbml_output, end=" ")
    # write_sbml_model(cs_model, sbml_output, use_fbc_package=args.use_fbc)
    write_sbml_model(cs_model, sbml_output)
    print("OK!")

    if args.write_confidences:
        csv_output = generate_output_fname(args, 'tsv')
        df_rxn_conf = create_confidences(cs_model, reaction_confidences)
        print("Writing reaction confidences to %s:" % csv_output, end=" ")
        df_rxn_conf.to_csv(csv_output, sep='\t')
        print("OK!")


main()



#!/usr/bin/env python3
import argparse
import os

import pandas as pd

from cobra.io import read_sbml_model
from cobra.util import solvers
from cobra.flux_analysis import pfba
from cobra.flux_analysis import single_gene_deletion
from cobra.flux_analysis import double_gene_deletion

from csmemo.cmds.io import read_json
from csmemo.cmds.manipulate import set_medium
from csmemo.cmds.manipulate import add_atpm_reaction
from csmemo.consistency_analysis import UmFinder


EXPERIMENTS = ['test_biomass', 'test_atp_yields', 'test_met_func', 'fba', 'single_gene_ko',
               'double_gene_ko', 'minimal_media', 'consistency_analysis']

SOLVERS = list(solvers.keys())
if 'cglpk' in SOLVERS:
    SOLVERS.remove('cglpk')
    SOLVERS.append('glpk')


def create_parser():
    parser = argparse.ArgumentParser(description='Run an in-silico experiment genome-scale metabolic model.')
    parser.add_argument('sbml_fname', action="store", help='SBML file to use a the model reference')
    
    parser.add_argument('--media', action="store", dest="json_exchanges", default=None,
                        help='JSON file storing the exchange bounds')

    parser.add_argument('--experiment', action="store", dest="experiment", choices=EXPERIMENTS,
                        required=True, help='Choose the in-silico experiment to be performed')

    parser.add_argument('--json-yields', action="store", dest="json_yields",
                        help='Path to json file with ATP yield reference')

    parser.add_argument('--json-met-func', action="store", dest="json_met_func",
                        help='Path to json file with metabolic functions')

    parser.add_argument('--solver', action="store", dest="solver", choices=SOLVERS,
                        default='glpk', help='LP solver to perform optimizations')

    parser.add_argument('--out', action="store", dest="output_folder", default=".",
                        help='Output folder to store the builded CSM')

    parser.add_argument('--processes', action="store", dest="processes", default=1, type=int,
                        help='Number of processesors used for computation')

    return parser


def generate_output_fname(output_folder, model_id, experiment, solver, output_format='tsv'):
    fname = "_".join([model_id, experiment, solver])
    fname = ".".join([fname, output_format])
    fname = os.path.join(output_folder, fname)
    return fname


def test_biomass_production(model, biomass_id='biomass_reaction'):
    # PREPARING MODEL FOR COMPONENTS PRODUCTION TEST
    model = model.copy()
    if 'ATPM' not in model.reactions:
        add_atpm_reaction(model)

    biomass_reaction = model.reactions.get_by_id(biomass_id)
    for r in model.reactions:
        if r.objective_coefficient == 0:
            continue
        r.objective_coefficient = 0

    biomass_components = dict()
    biomass_components['Energy'] = model.reactions.ATPM
    biomass_components['Biomass'] = model.reactions.biomass_reaction
    for m in biomass_reaction.reactants:
        reaction_list = list(m.reactions)
        reaction_list.remove(biomass_reaction)
        biomass_components[m.id] = reaction_list.pop()
        model.add_boundary(m, type='demand', lb=0.0)

    # The individual biomass components:
    # biomass_DNA_c, biomass_RNA_c, biomass_carbohydrate_c,
    # biomass_lipid_c, biomass_protein_c, biomass_other_c
    # Energy (ATPM) & biomass_reaction
    index = sorted(biomass_components.keys())
    df_result = pd.DataFrame(0.0, index=index, columns=['fOpt'])
    df_result.index.name = 'Component'
    for component, demand_rxn in biomass_components.items():
        demand_rxn.objective_coefficient = 1
        solution = model.optimize()
        demand_rxn.objective_coefficient = 0
        df_result.at[component, 'fOpt'] = round(solution.f, 4)

    return df_result


def test_atp_yields(model, atp_yields, zero_cutoff=1e-7, oxygen_bound=20):
    model = model.copy()
    if 'ATPM' not in model.reactions:
        add_atpm_reaction(model)

    for r in model.reactions:
        if r.objective_coefficient == 0:
            continue
        r.objective_coefficient = 0
    
    model.reactions.ATPM.objective_coefficient = 1

    index = sorted(atp_yields.keys())
    columns = ['Predicted', 'Reference', 'Valid']
    df_result = pd.DataFrame(index=index, columns=columns)
    df_result.index.name = 'rxn_id'

    for aminoacid_exchange, amino_acid_atp_yield in atp_yields.items():
        the_medium = {aminoacid_exchange: 1, 'EX_o2_e': oxygen_bound}
        set_medium(model, the_medium, inplace=True)
        solution = pfba(model)

        valid = abs(solution.fluxes.ATPM - amino_acid_atp_yield) < zero_cutoff
        df_result.at[aminoacid_exchange, 'Predicted'] = solution.fluxes.ATPM
        df_result.at[aminoacid_exchange, 'Reference'] = amino_acid_atp_yield
        df_result.at[aminoacid_exchange, 'Valid'] = valid

    return df_result


def test_metabolic_functions(model, met_func, zero_cutoff=1e-7):
    model = model.copy()
    met_func_dict = {}
    for m in met_func:
        if m in model.metabolites:
            try:
                met = model.metabolites.get_by_id(m)
                met_func_dict[m] = model.add_boundary(met, type='demand')
            except:
                print("Demand reactions already present for %s" % m)
                demand_id = "DM_" + m
                met_func_dict[m] = model.reactions.get_by_id(demand_id)

        else:
            met_func_dict[m] = None
    
    for r in model.reactions:
        if r.objective_coefficient == 0:
            continue
        r.objective_coefficient = 0
    
    index = sorted(met_func_dict.keys())
    columns = ['fluxes', 'produced']
    df_result = pd.DataFrame(index=index, columns=columns)
    df_result.index.name = "met_id"
    
    for idx, rxn in met_func_dict.items():
        if rxn:
            rxn.objective_coefficient = 1
            sol = model.optimize()
            df_result.at[idx, 'fluxes'] = sol.f
            df_result.at[idx, 'produced'] = sol.f > zero_cutoff
            rxn.objective_coefficient = 0
        else:
            df_result.at[idx, 'fluxes'] = 0
            df_result.at[idx, 'produced'] = False
    
    return df_result
    

def run_insilico_experiment(model, args, rescale=False):

    df_results = None

    if args.experiment == 'test_atp_yields':
        assert args.json_yields
        yields = read_json(args.json_yields)
        df_results = test_atp_yields(model, yields)
    
    if args.experiment == 'test_met_func':
        print("Reading Metabolic Functions: %s:" % args.json_met_func, end=" ")
        met_func = read_json(args.json_met_func)
        print("OK!")
        df_results = test_metabolic_functions(model, met_func)

    if args.experiment == 'test_biomass':
        df_results = test_biomass_production(model)

    if args.experiment == 'fba':
        solution = pfba(model)
        df_results = solution.to_frame()
        df_results.index.name = 'rxn_id'

    if args.experiment == 'single_gene_ko':
        if rescale:
            for r in model.exchanges:
                r.lower_bound *= 10
                r.upper_bound *= 10
        df_results = single_gene_deletion(model, processes=args.processes)

    if args.experiment == 'double_gene_ko':
        if rescale:
            for r in model.exchanges:
                r.lower_bound *= 10
                r.upper_bound *= 10
        
        genes = model.genes
        df_results = double_gene_deletion(model, gene_list1=genes,
                                          gene_list2=genes, processes=args.processes)

    if args.experiment == 'minimal_media':
        raise NotImplementedError
        # df_results = find_minimal_media(model)

    if args.experiment == 'consistency_analysis':
        gap_finder = UmFinder(model)
        blk_rxns = gap_finder.blocked_reactions
        gap_mets = gap_finder.gap_metabolites
        df_results = gap_finder.unconnected_modules_frame()

    return df_results


def main():

    parser = create_parser()
    args = parser.parse_args()

    assert os.path.isfile(args.sbml_fname)
    assert os.path.isdir(args.output_folder)
    assert os.path.isdir(args.output_folder)

    # Reading SBML genome-scale model
    print("Reading SBML Model from %s:" % args.sbml_fname, end=" ")
    model = read_sbml_model(args.sbml_fname)
    model.solver = args.solver
    print("OK!")

    if args.json_exchanges:
        assert os.path.isfile(args.json_exchanges)
        print("Reading exchange fluxes bounds: %s:" % args.json_exchanges, end=" ")
        media_dict = read_json(args.json_exchanges)
        print("OK!")
        
        print("Setting exchange fluxes bounds:", end=" ")
        set_medium(model, media_dict, inplace=True)
        print("OK!")

    print("Running in-silico experiment: \"%s\":" % args.experiment)
    df_result = run_insilico_experiment(model, args)
    print("OK!")

    tsv_output = generate_output_fname(args.output_folder, model.id,
                                       args.experiment, args.solver)
    
    print("Writing experiment results to %s:" % tsv_output, end=" ")
    df_result.to_csv(tsv_output, sep='\t')
    print("OK!")


main()



#!/usr/bin/env python3
import argparse
import os

from cobra.flux_analysis import find_blocked_reactions
from cobra.io import read_sbml_model
from cobra.io import write_sbml_model

from csmemo.cmds.io import read_json
from csmemo.cmds.manipulate import set_medium


def create_parser():
    
    parser = argparse.ArgumentParser(description="Trim blocked reactions and gap metabolites from \
                                                  a genome-scale metabolic model")
    
    parser.add_argument('sbml_fname', action="store", help='SBML file to use a the model reference')
    
    parser.add_argument('--out', action="store", dest="sbml_fname_out", required=True,
                         help='SBML file name to for the outputed model')
    
    parser.add_argument('--media', action="store", dest="json_exchanges", default=None,
                         help='JSON file storing the exchange bounds')
    
    parser.add_argument('--open-exchanges', action="store_true", dest="open_exchanges",
                         help="A flag to indicade wheather to relax exchange fluxes bounds. \
                               Ignored if --media is also used")
    
    parser.add_argument('--exchange-prefix', action="store", dest="exchange_prefix", default="EX",
                         help='Prefix for the exchange reaction. Use with open-exhanges')
                         
    parser.add_argument('--flux-bound', action="store", dest="flux_bound", default=1000.,
                         help='Prefix for the exchange reaction. Use with open-exhanges')
                         
    parser.add_argument('--usefbc', action="store_true", help='Write SBML files using FBC package')
   
    return parser


def main():

    parser = create_parser()
    args = parser.parse_args()

    assert os.path.isfile(args.sbml_fname)
        
    # Reading SBML genome-scale model
    print("Reading SBML Model from %s:" % args.sbml_fname, end=" ")
    model = read_sbml_model(args.sbml_fname)
    print("OK!")
    
    if args.json_exchanges:
        print("Reading exchange fluxes bounds: %s:" % args.json_exchanges, end=" ")
        media_dict = read_json(args.json_exchanges)
        print("OK!")
        
        print("Setting exchange fluxes bounds")
        set_medium(model, media_dict, inplace=True)
        print("OK!")
    else:
        if args.open_exchanges:
            for r in model.reactions:
                if not r.id.startswith(args.exchange_prefix):
                    continue
                r.lower_bound = -args.flux_bound
                r.upper_bound = args.flux_bound

    print("Finding blocked reactions and gap metabolites:", end=" ")
    blocked = find_blocked_reactions(model)
    blocked = set(blocked)
    gap_metabolites = [m for m in model.metabolites
                        if len(set([r.id for r in m.reactions]) - blocked) == 0]
    print("OK!")
    
    if len(blocked) > 0:
        print("- %i blocked reactions found" % len(blocked))
        print("- %i gap metabolites found" % len(gap_metabolites))
        print("Trimming model", end=" ")
        model.remove_reactions(blocked)
        model.remove_metabolites(gap_metabolites)
        print("OK!")
        print("Writing trimmed model as %s" % args.sbml_fname_out, end=" ")
        write_sbml_model(model, args.sbml_fname_out, use_fbc_package=args.usefbc)
        print("OK!")
    else:
        print("NO blocked reactions found, nothing to do")
    

main()



import warnings
from cobra.core import Reaction
from cobra.core import Metabolite


def set_medium(model, medium, skip_absent=True, inplace=False):
    if not inplace:
        model = model.copy()

    def set_active_bound(reaction, bound):
        if reaction.reactants:
            reaction.lower_bound = -bound
        elif reaction.products:
            reaction.upper_bound = bound

    # Set the given media bounds
    media_rxns = list()
    for rxn_id, bound in medium.items():
        if rxn_id not in model.reactions and skip_absent:
            warnings.warn("Exchange flux %s not found, skippied" % rxn_id)
            continue
        rxn = model.reactions.get_by_id(rxn_id)
        media_rxns.append(rxn)
        set_active_bound(rxn, bound)

    boundary_rxns = set([r for r in model.exchanges if r.id.startswith('EX')])
    media_rxns = set(media_rxns)

    # Turn off reactions not present in media
    for rxn in (boundary_rxns - media_rxns):
        set_active_bound(rxn, 0)

    if not inplace:
        return model


def add_reaction_dict(model, reaction_id, reaction_dict, compartment='c', lb=0.0, ub=1000, replace=False, inplace=True):
    
    if not inplace:
        model = model.copy()

    metabolites_to_add = {}
    for met_id, coeff in reaction_dict.items():
        if len(compartment) > 0:
            met_id += "_" + compartment
        
        try:
            metabolite = model.metabolites.get_by_id(met_id)
        except:
            warnings.warn("Metabolites %s no present in model, added as new" % met_id)
            metabolite = Metabolite(met_id)

        metabolites_to_add[metabolite] = coeff

    if reaction_id in model.reactions:
        if replace:
            model.remove_reactions([reaction_id])
        else:
            return
    
    reaction = Reaction(reaction_id)
    reaction.lower_bound = lb
    reaction.upper_bound = ub
    reaction.add_metabolites(metabolites_to_add)

    model.add_reaction(reaction)
    
    if not inplace:
        return model


def add_atpm_reaction(model, **kwargs):
    metabolites_dict = {"atp": -1, "h2o": -1, "adp": 1, "pi": 1, "h": 1}
    rxn_id = 'ATPM'
    add_reaction_dict(model, rxn_id, metabolites_dict, kwargs)




#!/usr/bin/env python
# coding: utf-8

from itertools import tee
import json
import argparse
import logging
from os import path

from cobra.io import read_sbml_model
from cobra.util import solvers

from csmemo.reachability import ReachabilityTester
from csmemo.consistency_analysis import get_unconnected_modules_frame


SOLVERS = list(solvers.keys())
if 'cglpk' in SOLVERS:
    SOLVERS.remove('cglpk')
    SOLVERS.append('glpk')

METHODS = ('lp7', 'fva')
DELETION_TYPES = ('gene', 'reaction')


def create_parser():
    parser = argparse.ArgumentParser(description='Run reachability analysis on genome-scale metabolic model.')

    parser.add_argument('sbml_fname', action="store", help='SBML file to use a the model reference')

    parser.add_argument('compounds_fname', action="store", help='JSON file storing the id of the metabolite to test')

    parser.add_argument('--type', action="store", dest="deletion_type", choices=DELETION_TYPES,
                        default='gene', help='Set the element type (gene/reaction) to test')

    parser.add_argument('--media', action="store", dest="json_exchanges", default=None,
                        help='JSON file storing the exchange bounds')

    parser.add_argument('--method', action="store", dest="method", choices=METHODS,
                        default='lp7', help='Optimization method to perform reachability')

    parser.add_argument('--solver', action="store", dest="solver", choices=SOLVERS,
                        default='glpk', help='LP solver to perform optimizations')

    parser.add_argument('--out', action="store", dest="output_folder", default=".",
                        help='Output folder to store the builded CSM')

    parser.add_argument('--verbose', action="store_true", dest="verbose", default=False,
                        help='Verbose mode')

    parser.add_argument('--find-ums', action="store_true", dest="find_ums", default=False,
                        help='Calculate unconnected modules set for blocked precursors')

    parser.add_argument('--save-square', action="store_true", dest="save_square", default=False,
                        help='Save the reachability matrix in dense form (rectangular matrix)')


    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()

    level = logging.WARNING
    if args.verbose:
        level = logging.INFO

    print("===================================================================")
    print("Reading sbml model: %s -->" % args.sbml_fname, end=" ")
    model = read_sbml_model(args.sbml_fname)
    model.solver = args.solver
    print("OK!")

    if args.json_exchanges:
        print("Reading media composition (json file): %s -->" % args.json_exchanges, end=" ")
        with open(args.json_exchanges) as fh:
            media_dict = json.load(fh)
            print("OK!")

        print("Setting media to model", end=" ")
        media_dict = {k: v for k, v in media_dict.items() if k in model.reactions}
        for r in model.exchanges:
            r.lower_bound = 0
            if r.id in media_dict:
                print(r.id)
                r.lower_bound = -1 * media_dict[r.id]

        print("OK!")

    print("Reading compound list (json file): %s -->" % args.compounds_fname, end=" ")
    with open(args.compounds_fname) as fh:
        all_compounds = json.load(fh)
    print("OK!")

    compounds_in_model = [i for i in all_compounds if i in model.metabolites]
    excluded_compounds = set(all_compounds) - set(compounds_in_model)

    reachability_tester = ReachabilityTester(model, compounds_in_model)
    default_not_produced = reachability_tester.run_reactions_reachability([])

    print("==========================================================================")
    print("- Initial set of biomass precursors to test: %i" % len(all_compounds))
    print("- Excluded compounds (not present in model) %i:" % len(excluded_compounds))
    print(" ".join(excluded_compounds))
    print("- Excluded compounds (not produced by model) %i:" % len(default_not_produced))
    print(" ".join(default_not_produced))
    print("- Total biomass precursors to be tested: %i" % len(compounds_in_model))
    print("==========================================================================")

    compounds_in_model = [i for i in compounds_in_model if i not in default_not_produced]



    reachability_tester = ReachabilityTester(model, compounds_in_model, level=level)

    print("Running single % reachability using method %s --> " % (args.deletion_type, args.method), end=" ")
    if args.deletion_type == 'gene':
        df_reachability = reachability_tester.single_gene_reachability(method=args.method)

    elif args.deletion_type == 'reaction':
        df_reachability = reachability_tester.single_reaction_reachability(method=args.method)

    print("OK!")

    column_id = args.deletion_type + "_id"
    elements = df_reachability[column_id].unique()

    compound_col = "compound_id"
    compounds = df_reachability[compound_col].unique()
    print("===================================================================")
    print("- %s disabling compound biosynthesis %i:" % (args.deletion_type, len(elements)))
    print(" ".join(sorted(elements)))
    print("- Compound no reached by at least one %s knockout %i:" % (args.deletion_type,len(compounds)))
    print(" ".join(sorted(compounds)))
    print("===================================================================")

    df_unconnected_modules = None
    if args.find_ums:
        print("Calculating unconnected_modules", end=" ")
        df_unconnected_modules = get_unconnected_modules_frame(model, elements, element_type=args.deletion_type)
        print("OK!")

    fname = path.join(args.output_folder, "single_" + args.deletion_type + "_reachability_" + model.id + ".tsv")
    print("Saving reachability frame to %s" % fname)
    df_reachability.to_csv(fname, sep='\t')
    
    if args.save_square:
        print("Saving reachability square frame to %s" % fname)
        fname = path.join(args.output_folder, "single_" + args.deletion_type + "_reachability_square_" + model.id + ".tsv")
        df_reachability_square = reachability_tester.to_square_form(df_reachability)
        df_reachability_square.to_csv(fname, sep='\t')

    if df_unconnected_modules is not None:
        fname = path.join(args.output_folder, "single_" + args.deletion_type + "_reachability_" + model.id + ".tsv")
        print("Saving reachability unconnected modules frame to %s" % fname)
        df_unconnected_modules.to_csv(fname, sep='\t')



if __name__ == "__main__":
    main()




import networkx as nx
import pandas as pd
from cobra.flux_analysis import find_blocked_reactions
from csmemo.reachability import find_gene_knockout_reactions


class UmFinder:
    def __init__(self, cobra_model, cc_method='fva', report=True):
        self._model = cobra_model

        if report:
            print("===========================================================")
            print("Initializing UmFinder Builder using")
            print("Model: %s" % cobra_model.id)

            print("- Nº of reactions: %i" % len(self._model.reactions))
            print("- Nº of metabolites: %i" % len(self._model.metabolites))

            print("\nChecking network consistency (may take some minutes)")
            print("Finding blocked reaction using method: %s\n" % cc_method)

        self._blocked_reactions = find_blocked_reactions(self.model)
        self._gap_metabolites = UmFinder.find_gap_metabolites(self.model, self.blocked_reactions)
        self._gap_graph = UmFinder.create_gap_graph(self.model, self._gap_metabolites, self._blocked_reactions)

        unconnected_modules = nx.connected_components(self._gap_graph.to_undirected())
        self._unconnected_modules = sorted(unconnected_modules, key=lambda x: len(x), reverse=True)

        if report:
            print("- Nº of blocked reactions: %i" % len(self._blocked_reactions))
            print("- Nº of gap metabolites: %i" % len(self._gap_metabolites))
            print("- Nº of unconnected modules: %i" % len(self.unconnected_modules))

            if len(self.unconnected_modules):
                df_ums = self.unconnected_modules_frame()
                df_biggest_um = df_ums.node_type[df_ums.um_id == 1]
                rxns = df_biggest_um.index[df_biggest_um =='rxn']
                mets = df_biggest_um.index[df_biggest_um == 'met']
                print("- N of reactions in the biggest unconnected module: %i" % len(rxns))
                print("- N of metabolites in the biggest unconnected module: %i" % len(mets))


    @property
    def model(self):
        return self._model

    @property
    def gap_metabolites(self):
        return frozenset(self._gap_metabolites)

    @property
    def gap_graph(self):
        return self._gap_graph

    @property
    def blocked_reactions(self):
        return frozenset(self._blocked_reactions)

    @property
    def unconnected_modules(self):
        return self._unconnected_modules

    def update(self):
        self._blocked_reactions = find_blocked_reactions(self.model)
        self._gap_metabolites = UmFinder.find_gap_metabolites(self.model, self.blocked_reactions)
        self._gap_graph = UmFinder.create_gap_graph(self.model, self._gap_metabolites, self._blocked_reactions)
        self._unconnected_modules = nx.connected_component_subgraphs(self._gap_graph.to_undirected())
        self._unconnected_modules = sorted(self._unconnected_modules, key=lambda x: len(x), reverse=True)

    def unconnected_module_subgraphs(self):
        for um in self.unconnected_modules:
            yield self.gap_graph.subgraph(um)

    def unconnected_modules_frame(self, unconnected_modules=[]):
        columns = ['node_id', 'node_type', 'um_id']
        data = {}
        counter = 0
        if len(unconnected_modules) == 0:
            unconnected_modules = self.unconnected_modules
        for i, um in enumerate(unconnected_modules):
            for e in um:
                if e in self.gap_metabolites:
                    e_type = 'met'
                elif e in self.blocked_reactions:
                    e_type = 'rxn'
                else:
                    e_type = None
                data[counter] = (e, e_type, i+1)
                counter += 1

        return pd.DataFrame.from_dict(data, orient='index', columns=columns)

    def find_new_ums(self, reactions_to_inactivate):
        if len(reactions_to_inactivate) == 0:
            return []

        if hasattr(reactions_to_inactivate[0], 'id'):
            reactions_to_inactivate = [r.id for r in reactions_to_inactivate]

        bounds = {}

        for r in reactions_to_inactivate:
            r = self.model.reactions.get_by_id(r)
            bounds[r.id] = r.bounds
            r.bounds = (0, 0)
            print(r)

        new_blocked_reactions = find_blocked_reactions(self.model)

        for r in reactions_to_inactivate:
            r = self.model.reactions.get_by_id(r)
            r.bounds = bounds[r.id]

        new_blocked_reactions = set(new_blocked_reactions) - self.blocked_reactions
        new_blocked_reactions = list(new_blocked_reactions)
        print(new_blocked_reactions)
        new_gap_metabolites = UmFinder.find_gap_metabolites(self.model, new_blocked_reactions)
        new_gap_graph = UmFinder.create_gap_graph(self.model, new_gap_metabolites, new_blocked_reactions)
        new_unconnected_modules = nx.connected_components(new_gap_graph.to_undirected())

        return new_unconnected_modules


    @staticmethod
    def find_gap_metabolites(model, blocked_reactions=[]):
        if len(blocked_reactions) == 0:
            blocked_reactions = find_blocked_reactions(model)
        gap_metabolites = []
        for m in model.metabolites:
            reactions = set([r.id for r in m.reactions])
            if reactions.issubset(blocked_reactions):
                gap_metabolites.append(m.id)

        return gap_metabolites

    @staticmethod
    def create_metabolic_graph(cobra_model, directed=True, reactions=None, rev_rxn_label='reversible'):

        graph = nx.DiGraph()
        if not directed:
            graph = nx.Graph()

        if not reactions:
            reactions = cobra_model.reactions

        if not hasattr(reactions[0], 'id'):
            reactions = [cobra_model.reactions.get_by_id(r) for r in reactions]

        for r in reactions:
            graph.add_node(r.id, label=r.id, text=r.id, node_class="rxn", node_id=r.id)
            for m in r.metabolites:
                if m.id not in graph.nodes():
                    graph.add_node(m.id, label=m.id, text=m.id, node_class="met", node_id=m.id)

                (tail, head) = (r.id, m.id)
                if r.get_coefficient(m) < 0:
                    (tail, head) = (m.id, r.id)

                graph.add_edge(tail, head)
                graph[tail][head][rev_rxn_label] = r.lower_bound < 0

        return graph

    @staticmethod
    def create_gap_graph(model, gap_metabolites, blocked_reactions):

        if len(gap_metabolites):
            if hasattr(gap_metabolites[0], 'id'):
                gap_metabolites = [m.id for m in gap_metabolites]

        if len(blocked_reactions) > 0:
            if hasattr(blocked_reactions[0], 'id'):
                blocked_reactions = [r.id for r in blocked_reactions]

        graph = UmFinder.create_metabolic_graph(model)
        gap_graph = graph.subgraph(gap_metabolites + blocked_reactions)

        return gap_graph


def get_unconnected_modules_frame(model, element_list, element_type='gene'):
    df = None
    um_finder = UmFinder(model)
    result_dict = {}
    if element_type == 'gene':
        for g in element_list:
            gene = model.genes.get_by_id(g)
            reactions_to_disable = find_gene_knockout_reactions(model, [gene])
            if len(reactions_to_disable) == 0:
                continue
            result_dict[g] = um_finder.find_new_ums(reactions_to_disable)

        data = []
        for g in result_dict:
            for i, um in enumerate(result_dict[g]):
                gene_id = "%s.%i" % (g, i+1)
                for e in um:
                    if e in model.reactions:
                        etype = 'rxn'
                    elif e in model.metabolites:
                        etype = 'met'
                    row = (gene_id, e, etype)
                    data.append(row)

        index = range(len(data))
        columns = ('gene_id', 'element_id', 'element_type')
        df = pd.DataFrame(data, columns=columns, index=index)

        return df

    if element_type == 'reactions':
        um_finder.find_new_ums(element_list)

    return df







import os
import setuptools
from setuptools import find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "csmemo",
    version = "0.0.6",
    author = "Miguel Ponce-de-Leon",
    author_email = "miguelponcedeleon@gmail.com",
    maintainer = "Miguel Ponce-de-Leon",
    maintainer_email = "miguelponcedeleon@gmail.com",
    description = "A Library implementing constraint-based methods for context-specific metabolic modeling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url = "https://github.com/bsc-life/csmemo",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires='>=3.6',
    setup_requires=['numpy', 'cobra>=0.25.0', 'corda', 'boolean.py', 'pyfastcore', 'networkx', 'pandas'],
    install_requires=['numpy', 'cobra>=0.25.0', 'corda', 'boolean.py', 'pyfastcore', 'networkx', 'pandas'],
    entry_points={
        'console_scripts': ['build-csm=csmemo.cmds.build_csm:main',
                            'run-experiment=csmemo.cmds.run_insilico_experiment:main',
                            'run-reachability=csmemo.cmds.run_reachability_analysis:main']
        
    }
)
